from gym.spaces.box import Box
from gym.spaces.discrete import Discrete
from gym.spaces.multi_discrete import MultiDiscrete
from gym.spaces.multi_binary import MultiBinary
from gym.spaces.prng import seed
from gym.spaces.tuple_space import Tuple

__all__ = ["Box", "Discrete", "MultiDiscrete", "MultiBinary", "Tuple"]
